# Unit tests for parsing and decoding datetime formats

from __future__ import annotations

import datetime

from pdfnaut.common.dates import encode_iso8601, encode_iso8824, parse_iso8601, parse_iso8824


def test_parse_iso8824() -> None:
    # Some examples from the spec
    assert parse_iso8824("D:199812231952-08'00") == datetime.datetime(
        1998, 12, 23, 19, 52, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=-8, minutes=0))
    )
    assert parse_iso8824("D:20010727133720") == datetime.datetime(
        2001, 7, 27, 13, 37, 20, tzinfo=datetime.timezone.utc
    )

    # Only date
    assert parse_iso8824("D:19981223Z") == datetime.datetime(
        1998, 12, 23, 0, 0, 0, tzinfo=datetime.timezone.utc
    )

    # Date string with ending apostrophe (pre 2.0)
    assert parse_iso8824("D:2024'") == datetime.datetime(
        2024, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc
    )


def test_encode_iso8824():
    # Date and time
    assert encode_iso8824(datetime.datetime(2001, 7, 27, 13, 37, 20)) == "D:20010727133720Z"

    # Only date
    assert encode_iso8824(datetime.datetime(2001, 7, 27), full=False) == "D:20010727"

    # With timezone
    tzd = datetime.timezone(datetime.timedelta(hours=-6, minutes=0))
    assert (
        encode_iso8824(datetime.datetime(2001, 7, 27, 13, 37, 20, tzinfo=tzd))
        == "D:20010727133720-06'00"
    )


def test_parse_iso8601() -> None:
    # Only year
    assert parse_iso8601("2025") == datetime.datetime(
        2025, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
    )
    # Date
    assert parse_iso8601("2025-02-01") == datetime.datetime(
        2025, 2, 1, 0, 0, tzinfo=datetime.timezone.utc
    )
    # Date and time without seconds
    assert parse_iso8601("2025-02-01T12:31") == datetime.datetime(
        2025, 2, 1, 12, 31, tzinfo=datetime.timezone.utc
    )
    # Date and time with seconds
    assert parse_iso8601("2025-02-01T12:31:17") == datetime.datetime(
        2025, 2, 1, 12, 31, 17, tzinfo=datetime.timezone.utc
    )
    # Date and time with seconds and fraction unit
    assert parse_iso8601("2025-02-01T12:31:17.20") == datetime.datetime(
        2025, 2, 1, 12, 31, 17, 200_000, tzinfo=datetime.timezone.utc
    )
    # Date and time with seconds, fraction unit, and timezone
    tzd = datetime.timezone(datetime.timedelta(hours=-6, minutes=0))
    assert parse_iso8601("2025-02-01T12:31:17.20-06:00") == datetime.datetime(
        2025, 2, 1, 12, 31, 17, 200_000, tzd
    )


def test_encode_iso8601():
    # Date and time
    assert encode_iso8601(datetime.datetime(2001, 7, 27, 13, 37, 20)) == "2001-07-27T13:37:20Z"

    # Datetime with timezone
    tzd = datetime.timezone(datetime.timedelta(hours=-6, minutes=0))
    assert (
        encode_iso8601(datetime.datetime(2025, 2, 1, 12, 31, 17, 200_000, tzd))
        == "2025-02-01T12:31:17.200000-06:00"
    )

    # Only date
    assert encode_iso8601(datetime.datetime(2024, 10, 20), full=False) == "2024-10-20"

    # Date and time (utc)
    assert encode_iso8601(datetime.datetime(2024, 10, 20, 15, 34, 10)) == "2024-10-20T15:34:10Z"



from __future__ import annotations

import hashlib
from collections.abc import Iterable
from datetime import time
from typing import TypeVar

from ..cos.objects.base import ObjectGetter, PdfHexString, PdfObject, PdfReference
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.objects.stream import PdfStream


def get_value_from_bytes(contents: PdfHexString | bytes) -> bytes:
    """Returns the decoded value of ``contents`` if it is an instance of
    :class:`.PdfHexString`, otherwise returns ``contents``."""
    return contents.value if isinstance(contents, PdfHexString) else contents


R = TypeVar("R")


def ensure_object(obj: PdfReference[R] | R) -> R:
    """Resolves ``obj`` to a direct object if ``obj`` is an instance of
    :class:`.PdfReference`. Otherwise, returns ``obj`` as is."""
    if isinstance(obj, PdfReference):
        return obj.get()

    return obj


def get_closest(values: Iterable[int], target: int) -> int:
    """Returns the integer in ``values`` closest to ``target``."""
    return min(values, key=lambda offset: abs(offset - target))


def generate_file_id(filename: str, content_size: int) -> PdfHexString:
    """Generates a file identifier using ``filename`` and ``content_size`` as
    described in § 14.4, "File identifiers".

    File identifiers are values that uniquely separate a revision of a document
    from another. The file identifier is generated using the same information
    specified in the standard, that is, the current time, the file path and
    the file size in bytes.
    """

    id_digest = hashlib.md5(time().isoformat("auto").encode())
    id_digest.update(filename.encode())
    id_digest.update(str(content_size).encode())

    return PdfHexString(id_digest.hexdigest().encode())


def renumber_references(
    root: PdfObject, resolver: ObjectGetter, start: int = 1
) -> tuple[PdfObject, dict[int, PdfObject]]:
    """Renumbers all references in ``root`` (including its nested objects) based on ``start``.

    Arguments:
        root (PdfDictionary, PdfArray, or PdfStream):
            The root object to renumber references in.

        resolver (ObjectGetter):
            The resolver attached to each new reference.

        start (int, optional):
            An integer from which to start numbering new references (by default, 1).

    Returns:
        A tuple of two items including the root object with the renumbered references and a
        mapping of reference numbers to items that can be written to the object store.

    .. warning::
        Because this function works recursively, ``root`` shall not contain cyclic
        references, that is, reference paths that may point back to ``root``.
    """
    references = {}

    def inner(obj: PdfObject) -> PdfObject:
        nonlocal start

        if isinstance(obj, PdfDictionary):
            for key, value in obj.data.items():
                obj.data[key] = inner(value)
        elif isinstance(obj, PdfStream):
            for key, value in obj.details.data.items():
                obj.details.data[key] = inner(value)
        elif isinstance(obj, PdfArray):
            for idx, value in enumerate(obj.data):
                obj.data[idx] = inner(value)
        elif isinstance(obj, PdfReference):
            referred = inner(obj.get())

            val = PdfReference(start, 0).with_resolver(resolver)
            start += 1

            references[val.object_number] = referred
            return val

        return obj

    return inner(root), references



from __future__ import annotations

import re
from collections.abc import Callable
from typing import cast

from .objects import (
    ObjectGetter,
    PdfArray,
    PdfComment,
    PdfDictionary,
    PdfHexString,
    PdfInlineImage,
    PdfName,
    PdfNull,
    PdfObject,
    PdfOperator,
    PdfReference,
)

# as defined in § 7.2.3, "Character Set", Table 1 & 2
DELIMITERS = b"()<>[]{}/%"
WHITESPACE = b"\x00\t\n\x0c\r "
EOL_CR = b"\r"
EOL_LF = b"\n"
EOL_CRLF = b"\r\n"

# as defined in § 7.3.4.2, "Literal Strings", Table 3
STRING_ESCAPE = {
    b"\\n": b"\n",
    b"\\r": b"\r",
    b"\\t": b"\t",
    b"\\b": b"\b",
    b"\\f": b"\f",
    b"\\(": b"(",
    b"\\)": b")",
    b"\\\\": b"\\",
}


class ContentStreamTokenizer:
    """A tokenizer designed to consume the contents within a content stream.

    This tokenizer relies on :class:`PdfTokenizer` to parse common tokens
    but has special handling for the operators inside a content stream.
    """

    def __init__(self, contents: bytes) -> None:
        self.contents = contents
        self.tokenizer = PdfTokenizer(contents)

    def __iter__(self):
        return self

    def __next__(self) -> PdfOperator:
        while not self.tokenizer.done:
            if (operator := self.get_next_token()) is not None and isinstance(
                operator, PdfOperator
            ):
                return operator

        raise StopIteration

    def get_next_token(self) -> PdfOperator | PdfComment | None:
        """Consumes the next token.

        The return value is either a :class:`.PdfOperator` or a :class:`.PdfComment`
        in case a token was consumed or `None` if the end of data has been reached.
        """
        if self.tokenizer.done:
            return

        operands = []
        while not self.tokenizer.done:
            if (tok := self.tokenizer.get_next_token(parse_references=False)) is not None:
                if isinstance(tok, PdfComment):
                    return tok

                operands.append(tok)
                continue
            elif (pk := self.tokenizer.peek()).isalpha() or pk in b"'\"":
                name = self.tokenizer.consume_while(lambda ch: ch not in DELIMITERS + WHITESPACE)

                if name == b"BI":
                    # inline images must be handled specially so as to not
                    # confuse the parser.
                    return self.parse_inline_image()

                return PdfOperator(name, operands)

            self.tokenizer.skip()

    def parse_inline_image(self) -> PdfOperator:
        """Parses an inline image.

        Inline images are an alternative to image XObjects designed for embedding
        small images in a content stream.

        Returns an operator ``EI`` (for "end image") with a :class:`.PdfInlineImage`
        as its first and only operand.
        """
        mapping = self.tokenizer.parse_kv_map_until(b"ID")

        # Abbreviated names are preferred according to
        # https://github.com/pdf-association/pdf-issues/issues/3
        filter_names = mapping.get("F", mapping.get("Filter"))
        if filter_names is None:
            filter_names = PdfArray()

        if isinstance(filter_names, PdfName):
            filter_names = PdfArray([filter_names])

        filter_names = cast(PdfArray[PdfName], filter_names)

        # If the next character is whitespace, consume it.
        if self.tokenizer.peek() in WHITESPACE:
            self.tokenizer.consume()

        # However, if the filter is ASCIIHex or ASCII85, consume all of the whitespace
        # (including comments).
        checking_filters = (b"A85", b"AHx", b"ASCIIHexDecode", b"ASCII85Decode")

        if any(fn.value in checking_filters for fn in filter_names):
            self.tokenizer.skip_whitespace()
            self.tokenizer.skip_if_comment()

        # TODO: handle PDF 2.0's /L & /Length for inline images
        image_data = self.tokenizer.consume_while(lambda _: not self.tokenizer.peek(2) == b"EI")

        return PdfOperator(self.tokenizer.consume(2), [PdfInlineImage(mapping, image_data)])


class PdfTokenizer:
    """A tokenizer designed to consume individual objects that do not depend on a cross
    reference table. It is used by :class:`~pdfnaut.cos.parser.PdfParser` for this purpose.

    This tokenizer consumes basic objects such as arrays and dictionaries. Indirect objects
    and streams depend on an XRef table and hence are not sequentially parsable. It is not
    intended to parse these items but rather the objects stored within them.

    Arguments:
        data (bytes):
            The contents to be parsed.
    """

    def __init__(self, data: bytes) -> None:
        self.data = data
        self.position = 0
        self.resolver: ObjectGetter | None = None

    def __iter__(self):
        return self

    def __next__(self) -> PdfObject | PdfComment | PdfOperator:
        while not self.done:
            if (tok := self.get_next_token()) is not None:
                return tok
            self.skip()
        raise StopIteration

    # * Scanning
    @property
    def done(self) -> bool:
        """Whether the parser has reached the end of data."""
        return self.position >= len(self.data)

    def skip(self, n: int = 1) -> None:
        """Skips/advances ``n`` characters in the tokenizer."""
        if not self.done:
            self.position += n

    def peek(self, n: int = 1) -> bytes:
        """Peeks ``n`` characters into ``data`` without advancing through the tokenizer."""
        return self.data[self.position : self.position + n]

    def peek_line(self) -> bytes:
        """Peeks from the current position until an EOL marker is found (not included
        in the output)."""
        start_pos = self.position
        line = self.consume_while(lambda _: not self.peek(2).startswith((EOL_CRLF, EOL_CR, EOL_LF)))
        self.position = start_pos
        return line

    def consume(self, n: int = 1) -> bytes:
        """Consumes and returns ``n`` characters."""
        consumed = self.peek(n)
        self.skip(len(consumed))

        return consumed

    def matches(self, keyword: bytes) -> bool:
        """Checks whether ``keyword`` starts at the current position."""
        return self.peek(len(keyword)) == keyword

    def _get_reference_if_matched(self) -> re.Match[bytes] | None:
        """Returns the Regex match if an indirect reference is at the current position
        otherwise None."""
        if not self.peek().isdigit():
            return

        return re.match(rb"^(?P<num>\d+)\s+(?P<gen>\d+)\s+R", self.peek_line())

    def _is_octal(self, byte: bytes) -> bool:
        """Returns whether ``byte`` is a valid octal number (0-7)."""
        return b"0" <= byte <= b"7"

    def skip_if_matches(self, keyword: bytes) -> bool:
        """Advances ``len(keyword)`` characters if ``keyword`` starts at the current
        position. Returns whether the match was successful."""
        if self.matches(keyword):
            self.skip(len(keyword))
            return True
        return False

    def skip_if_comment(self) -> bool:
        """Advances through a PDF comment in case one occurs at the current position."""
        if self.matches(b"%"):
            self.parse_comment()
            return True
        return False

    def skip_whitespace(self) -> None:
        """Advances through PDF whitespace."""
        self.skip_while(lambda ch: ch in WHITESPACE)

    def skip_next_eol(self, no_cr: bool = False) -> None:
        """Skips the next EOL marker if matched. If ``no_cr`` is True, CR (``\\r``) as is
        will not be treated as a newline."""
        matched = self.skip_if_matches(EOL_CRLF)
        if no_cr and self.matches(EOL_CR):
            return

        if not matched and self.peek() in EOL_CRLF:
            self.skip()

    def skip_while(self, callback: Callable[[bytes], bool], *, limit: int = -1) -> int:
        """Skips while ``callback`` returns True for an input character. If specified,
        it will only skip ``limit`` characters. Returns how many characters were skipped."""
        if limit == -1:
            limit = len(self.data)

        start = self.position
        while not self.done and callback(self.peek()) and self.position - start < limit:
            self.position += 1
        return self.position - start

    def consume_while(self, callback: Callable[[bytes], bool], *, limit: int = -1) -> bytes:
        """Consumes while ``callback`` returns True for an input character. If specified,
        it will only consume up to ``limit`` characters."""
        if limit == -1:
            limit = len(self.data)

        consumed = b""
        while not self.done and callback(self.peek()) and len(consumed) < limit:
            consumed += self.consume()
        return consumed

    def get_next_token(self, *, parse_references: bool = True) -> PdfObject | PdfComment | None:
        """Parses and returns the token at the current position.

        Arguments:
            parse_references (bool, optional, keyword only):
                Whether to parse indirect references. This is intended for
                content streams where indirect references are disallowed.
        """
        if self.done:
            return

        if self.skip_if_matches(b"true"):
            return True
        elif self.skip_if_matches(b"false"):
            return False
        elif self.skip_if_matches(b"null"):
            return PdfNull()
        elif parse_references and (mat := self._get_reference_if_matched()):
            return self.parse_indirect_reference(mat)
        elif self.peek().isdigit() or self.peek() in b".+-":
            return self.parse_numeric()
        elif self.matches(b"["):
            return self.parse_array()
        elif self.matches(b"/"):
            return self.parse_name()
        elif self.matches(b"<<"):
            return self.parse_dictionary()
        elif self.matches(b"<"):
            return self.parse_hex_string()
        elif self.matches(b"("):
            return self.parse_literal_string()
        elif self.matches(b"%"):
            return self.parse_comment()

    def parse_numeric(self) -> int | float:
        """Parses a numeric object.

        PDF has two types of numbers: integers (40, -30) and real numbers (3.14). The range
        and precision of these numbers may depend on the machine used to process the PDF.
        """
        prefix_or_digit = self.consume()  # either a digit, a dot, or a sign prefix
        number = prefix_or_digit + self.consume_while(lambda ch: ch.isdigit() or ch == b".")

        # is this a float (a real number)?
        if b"." in number:
            return float(number)
        return int(number)

    def parse_name(self) -> PdfName:
        """Parses a name -- a uniquely defined atomic symbol introduced with a slash
        and ending before a delimiter or whitespace."""
        self.skip()  # past the /

        atom = b""
        while not self.done and self.peek() not in DELIMITERS + WHITESPACE:
            if self.matches(b"#"):
                # escape sequence matched
                self.skip()

                atom += int(self.consume(2), 16).to_bytes(1, "little")
                continue

            atom += self.consume()

        return PdfName(atom)

    def parse_hex_string(self) -> PdfHexString:
        """Parses a hexadecimal string. Hexadecimal strings usually include arbitrary binary
        data. If the sequence is uneven, the last character is assumed to be 0."""
        self.skip()  # adv. past the <

        content = self.consume_while(lambda ch: ch != b">")
        self.skip()  # adv. past the >

        return PdfHexString(content)

    def parse_dictionary(self) -> PdfDictionary:
        """Parses a dictionary object.

        In a PDF, dictionary keys are name objects and dictionary values are any
        object or reference. This parser maps name objects to strings in this
        context.
        """

        self.skip(2)  # adv. past the <<
        return self.parse_kv_map_until(b">>")

    def parse_kv_map_until(self, delimiter: bytes) -> PdfDictionary:
        """Parses from the current position a dictionary-like object,
        that is, an object composed of keys that are name objects and values
        that are any object.

        The 'delimiter' parameter specifies where this dictionary should end.
        The common ending (and default value) is ">>" for dictionary objects.
        However, this also accommodates for inline images which have the ID
        operator that can be used as a delimiter.
        """

        kv_pairs: list[PdfObject] = []

        while not self.done and not self.matches(delimiter):
            if (token := self.get_next_token()) is not None and not isinstance(token, PdfComment):
                kv_pairs.append(cast(PdfObject, token))

            # Only advance when no token matches. The individual object
            # parsers already advance and this avoids advancing past delimiters.
            if token is None:
                self.skip()

        self.skip(len(delimiter))

        return PdfDictionary(
            {
                cast(PdfName, kv_pairs[i]).value.decode(): kv_pairs[i + 1]
                for i in range(0, len(kv_pairs), 2)
            }
        )

    def parse_array(self) -> PdfArray:
        """Parses an array. Arrays are heterogenous in PDF so they are mapped to Python lists."""
        self.skip()  # past the [

        items = PdfArray[PdfObject]()

        while not self.done and not self.matches(b"]"):
            if (token := self.get_next_token()) is not None and not isinstance(token, PdfComment):
                items.append(cast(PdfObject, token))

            if token is None:
                self.skip()

        self.skip()  # past the ]

        return items

    def parse_indirect_reference(self, mat: re.Match[bytes]) -> PdfReference:
        """Parses an indirect reference. Indirect references allow locating an object in a PDF."""
        self.skip(mat.end())  # consume the reference
        self.skip_whitespace()

        reference = PdfReference(int(mat.group("num")), int(mat.group("gen")))
        if self.resolver:
            reference._resolver = self.resolver

        return reference

    def parse_literal_string(self) -> bytes:
        """Parses a literal string.

        Literal strings may be composed entirely of ASCII or may include arbitrary
        binary data. They may also include escape sequences and octal values (``\\ddd``).
        """
        self.skip()  # past the (

        string = b""
        # balanced parentheses do not require escaping
        paren_depth = 1

        while not self.done and paren_depth >= 1:
            if self.matches(b"\\"):
                # Is this a default escape? (Table 3 § 7.3.4.2)
                escape = STRING_ESCAPE.get(self.peek(2))

                if escape is not None:
                    string += escape
                    self.skip(2)  # past the escape code
                    continue

                # Otherwise, match a newline or a \ddd sequence
                self.skip(1)

                matched = self.skip_if_matches(EOL_CRLF)
                if not matched and self.peek() in EOL_CRLF:
                    self.skip()
                elif self._is_octal(self.peek()):
                    octal_code = self.consume_while(self._is_octal, limit=3)
                    # the octal value will be 8 bit at most
                    string += int(octal_code, 8).to_bytes(1, "little")
                    continue

            if self.matches(b"("):
                paren_depth += 1
            elif self.matches(b")"):
                paren_depth -= 1

            # This avoids appending the delimiting paren
            if paren_depth != 0:
                string += self.peek()

            self.skip()

        return string

    def parse_comment(self) -> PdfComment:
        """Parses a PDF comment. Comments have no syntactical meaning."""
        self.skip()  # past the %

        line = self.consume_while(lambda ch: ch not in EOL_CRLF)
        self.skip_whitespace()

        return PdfComment(line)



from __future__ import annotations


class PdfParseError(Exception):
    """The parser was unable to continue parsing the PDF."""

    pass


class PdfFilterError(PdfParseError):
    """A filter was unable to decode a stream or the filter in question is not
    supported."""

    pass


class PdfResolutionError(Exception):
    """The document was unable to resolve a reference because no resolution method
    is available."""

    pass


class PdfWriteError(Exception):
    """The writer was unable to serialize an object."""

    pass


class MissingCryptProviderError(Exception):
    """pdfnaut cannot encrypt or decrypt because no crypt provider is available."""

    pass



from __future__ import annotations

from collections.abc import Generator, Iterable, MutableSequence
from copy import deepcopy
from typing import Any, Iterator, cast, overload

from typing_extensions import Self

from .common.utils import renumber_references
from .cos.objects import PdfArray, PdfDictionary, PdfName, PdfReference
from .cos.parser import PdfParser
from .objects.page import Page


def flatten_pages(root: PdfDictionary) -> Generator[Page, None, None]:
    """Yields all :class:`.Page` objects within ``root`` and its descendants."""

    kids = cast(PdfArray, root["Kids"])

    for page_ref in cast(list[PdfReference], kids.data):
        page = cast(PdfDictionary, page_ref.get())

        type_ = cast(PdfName, page["Type"])
        if type_.value == b"Pages":
            yield from flatten_pages(page)
        elif type_.value == b"Page":
            yield Page.from_dict(page, indirect_ref=page_ref)


class PageList(MutableSequence[Page]):
    """A mutable sequence of the pages in a document.

    This class isn't designed to be constructed by a user. To access the page
    list of a PDF, use :class:`PdfDocument.pages`.
    """

    def __init__(
        self,
        pdf: PdfParser,
        root_tree: PdfDictionary,
        root_tree_ref: PdfReference,
    ) -> None:
        self._pdf = pdf
        self._root_tree = root_tree
        self._root_tree_ref = root_tree_ref
        self._indexed_page_cache = list(flatten_pages(self._root_tree))
        self._last_hash = hash(self._root_tree)

    def _update_on_hash(self) -> None:
        # process: if the page tree has changed, only replace the pages
        # in the indexed page cache that have also changed.

        if self._last_hash == hash(self._root_tree):
            return

        page_list: list[Page] = []

        for idx, page in enumerate(flatten_pages(self._root_tree)):
            if 0 <= idx < len(self._indexed_page_cache):
                # page in list, check if it is different.
                prev_page = self._indexed_page_cache[idx]
                if hash(prev_page) != hash(page):
                    page_list.append(page)
                else:
                    page_list.append(prev_page)
            else:
                # page not in list, simply append.
                page_list.append(page)

        self._indexed_page_cache = page_list

    def _get_indexed_pages(self) -> list[Page]:
        self._update_on_hash()
        return self._indexed_page_cache

    # * mutable sequence methods
    def __len__(self) -> int:
        return len(self._get_indexed_pages())

    def __contains__(self, value: object) -> bool:
        return value in self._get_indexed_pages()

    def __iter__(self) -> Iterator[Page]:
        return iter(self._get_indexed_pages())

    def __reversed__(self) -> Iterator[Page]:
        return reversed(self._get_indexed_pages())

    @overload
    def __getitem__(self, index: int) -> Page: ...

    @overload
    def __getitem__(self, index: slice) -> list[Page]: ...

    def __getitem__(self, index: int | slice) -> Page | list[Page]:
        return self._get_indexed_pages()[index]

    @overload
    def __setitem__(self, index: int, value: Page) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[Page]) -> None: ...

    def __setitem__(self, index: int | slice, value: Page | Iterable[Page]) -> None:
        if isinstance(index, slice):
            raise NotImplementedError

        assert isinstance(value, Page)

        result, _ = self._get_tree_with_index(self._root_tree, self._root_tree_ref, index)
        if result is None:
            raise IndexError("page tree assignment index out of range")

        tree, _, tree_idx = result

        value = self._add_page_to_obj_store(value)

        # delete the page being replaced from the object store
        replacing_ref = tree["Kids"].data[tree_idx]
        self._pdf.objects.delete(replacing_ref.object_number)
        self._indexed_page_cache[index].indirect_ref = None

        # set the page
        tree["Kids"][tree_idx] = value.indirect_ref
        self._indexed_page_cache[index] = value

    def __delitem__(self, index: int | slice) -> None:
        if isinstance(index, slice):
            raise NotImplementedError

        self.pop(index)

    def __iadd__(self, values: Iterable[Page]) -> Self:
        self.extend(values)
        return self

    def index(self, value: Any, start: int = 0, stop: int = ...) -> int:
        """Returns the index at which page ``value`` was first found in the
        range of ``start`` included to ``stop`` excluded."""
        return self._get_indexed_pages().index(value, start, stop)

    def count(self, value: Any) -> int:
        """Returns the amount of times page ``value`` appears in the page list."""
        return self._get_indexed_pages().count(value)

    def insert(self, index: int, value: Page) -> None:
        """Inserts a page ``value`` at ``index``. ``index`` is the index of
        the page before which to insert.

        When inserting, the page object is copied into the page list.

        Note: The identity of the output shall match the identity of the input page.
        The input page shall receive the indirect reference of the inserted page.
        """
        if index < -len(self):
            index = 0
        elif index >= len(self):
            return self._append_pages_to_tree([value])
        else:
            index = self._pos_idx_of(index)

        inserting_page = self._add_page_to_obj_store(value)

        if self._get_indexed_pages():
            # document has pages, traverse the tree and insert at location
            result, _ = self._get_tree_with_index(self._root_tree, self._root_tree_ref, index)
        else:
            # document has no pages, insert in root page tree
            result = (self._root_tree, self._root_tree_ref, index)

        # This should always be the case but, for good measure, we check it.
        assert result is not None, f"expected tree for index {index}."

        tree, tree_ref, tree_idx = result

        self._insert_page_into_tree(inserting_page, tree_idx, tree=tree, tree_ref=tree_ref)
        self._indexed_page_cache.insert(index, value)

    def append(self, value: Page) -> None:
        """Appends a page ``value`` to the page list."""
        self.insert(len(self._get_indexed_pages()), value)

    def clear(self) -> None:
        raise NotImplementedError

    def reverse(self) -> None:
        raise NotImplementedError

    def extend(self, values: Iterable[Page]) -> None:
        """Appends a list of pages ``values`` into the page list.

        When extending, all pages will be copied and inserted into the last page
        tree within the page list.

        Notes:
            - Resources not part of the page object, such as bookmarks or form
              objects, will not be preserved. You must add those yourself.
            - The indirect reference for each of the pages will only be changed if
              the pages have no references applied to them.
        """
        self._append_pages_to_tree(values)

    def pop(self, index: int = -1) -> Page:
        """Removes the page at ``index``.

        Only the page object is removed from the document. Additionally, its
        reference is also invalidated. The resources used by the page are left
        intact as they may be used later on in other pages.

        Raises:
            IndexError: The page list is empty or the index does not exist.

        Returns:
            Page: The page object that was popped.
        """
        index = self._pos_idx_of(index)

        if self._get_indexed_pages():
            # document has pages, traverse the tree and insert at location
            result, _ = self._get_tree_with_index(self._root_tree, self._root_tree_ref, index)
        else:
            result = None

        if result is not None:
            tree, _, tree_idx = result
        else:
            tree = self._root_tree
            tree_idx = index

        # delete the page from the tree
        self._delete_page_in_tree(tree_idx, tree)
        output = self._indexed_page_cache.pop(index)

        # delete the page from the object store
        if output.indirect_ref is not None:
            self._pdf.objects.delete(output.indirect_ref.object_number)
            output.indirect_ref = None

        return output

    def remove(self, value: Page) -> None:
        """Removes the first occurrence of page ``value`` in the document.

        Raises:
            IndexError: The page list is empty or the page is not in this document.
        """
        index = self.index(value)
        value.indirect_ref = None

        self.pop(index)

    # * helper methods
    def _pos_idx_of(self, index: int) -> int:
        # positive index is within 0 and len(self), both inclusive
        # if index < 0, index = len(self) - abs(index)

        if index >= 0:
            return min(index, len(self))

        return len(self) - abs(index)

    def _add_page_to_obj_store(self, page: Page) -> Page:
        # Ensure that the page has no parent. Having one will cause a
        # reference loop and the page will be re-parented anyways.
        if "Parent" in page:
            page.pop("Parent")

        if page.indirect_ref is not None:
            # page has an indirect ref, assume page comes from different
            # document and create copy.
            added_page, refs = renumber_references(
                PdfDictionary(deepcopy(page.data)),
                self._pdf.get_object,
                start=self._pdf.objects.get_next_ref().object_number,
            )

            for num, ref in refs.items():
                self._pdf.objects[num] = ref
        else:
            # no indirect reference, assume new page and create copy.
            added_page = PdfDictionary(deepcopy(page.data))

        page_ref = self._pdf.objects.add(added_page)
        added_page = Page.from_dict(added_page, indirect_ref=page_ref)

        # only set the reference if the page has none.
        if page.indirect_ref is None:
            page.indirect_ref = page_ref
            page.data = added_page.data
            return page

        return added_page

    def _insert_page_into_tree(
        self, page: Page, tree_index: int, *, tree: PdfDictionary, tree_ref: PdfReference
    ) -> None:
        if page.indirect_ref is None:
            raise ValueError("Page has no indirect reference assigned to it.")

        tree["Kids"].insert(tree_index, page.indirect_ref)
        tree["Count"] += 1

        page["Parent"] = tree_ref

        parent = tree
        while (parent := parent.get("Parent")) is not None:
            parent["Count"] += 1

    def _append_pages_to_tree(self, values: Iterable[Page]) -> None:
        last_tree, last_tree_ref = self._get_last_tree(self._root_tree, self._root_tree_ref)

        for value in values:
            inserting_page = self._add_page_to_obj_store(value)

            self._insert_page_into_tree(
                inserting_page, len(last_tree["Kids"]), tree=last_tree, tree_ref=last_tree_ref
            )
            self._indexed_page_cache.insert(len(self._indexed_page_cache), value)

    def _delete_page_in_tree(self, tree_index: int, tree: PdfDictionary) -> None:
        tree.data["Kids"].pop(tree_index)

        tree["Count"] -= 1

        parent = tree
        while (parent := parent.get("Parent")) is not None:
            parent["Count"] -= 1

    def _get_last_tree(
        self, root: PdfDictionary, root_ref: PdfReference
    ) -> tuple[PdfDictionary, PdfReference]:
        kids = cast(PdfArray[PdfReference], root["Kids"]).data
        result = (root, root_ref)

        for page_ref in kids:
            page = page_ref.get()
            type_ = cast(PdfName, page["Type"])

            if type_.value == b"Pages":
                result = self._get_last_tree(page, page_ref)

        return result

    def _get_tree_with_index(
        self, root: PdfDictionary, root_ref: PdfReference, index: int
    ) -> tuple[tuple[PdfDictionary, PdfReference, int] | None, int]:
        kids = cast(PdfArray[PdfReference], root["Kids"].data)

        for tree_index, page_ref in enumerate(kids):
            page = page_ref.get()

            type_ = cast(PdfName, page["Type"])

            if type_.value == b"Pages":  # intermediate node
                result, index = self._get_tree_with_index(page, page_ref, index)
                if result is not None:
                    return (result, index)
            elif type_.value == b"Page":  # page node
                if index <= 0:
                    return (root, root_ref, tree_index), index

                index -= 1

        return (None, index)



from __future__ import annotations

import enum
from typing import Literal

PageLayout = Literal[
    "SinglePage", "OneColumn", "TwoColumnLeft", "TwoColumnRight", "TwoPageLeft", "TwoPageRight"
]
PageMode = Literal["UseNone", "UseOutlines", "UseThumbs", "FullScreen", "UseOC", "UseAttachments"]


class UserAccessPermissions(enum.IntFlag):
    """User access permissions as specified in the P entry of the document's standard
    encryption dictionary. See "Table 22: User access permissions"."""

    PRINT = 1 << 2  # revision 3+, print may be limited by FAITHFUL_PRINT
    """Print the document. If the document uses revision 3 or later, print quality 
    may be influenced by :attr:`.FAITHFUL_PRINT`."""

    MODIFY = 1 << 3  # limited by MANAGE_ANNOTATIONS, FILL_FORM_FIELDS, and ASSEMBLE_DOCUMENT
    """Modify the contents of the document. May be influenced by :attr:`.MANAGE_ANNOTATIONS`, 
    :attr:`.FILL_FORM_FIELDS`, and :attr:`.ASSEMBLE_DOCUMENT`."""

    COPY_CONTENT = 1 << 4  # assumed as 1 by assistive technology
    """Copy or extract text and graphics. Assistive technology may assume
    this bit as 1 for its purposes, as per :attr:`.ACCESSIBILITY`."""

    MANAGE_ANNOTATIONS = 1 << 5  # affects forms too, affected by MODIFY
    """Add or modify text annotations, fill interactive form fields. And, depending 
    on whether :attr:`.MODIFY` is set, create and modify form fields."""

    FILL_FORM_FIELDS = 1 << 8  # revision 3+, overrides MODIFY
    """For documents using revision 3 or later, fill existing interactive form 
    fields, even if :attr:`.MANAGE_ANNOTATIONS` is clear."""

    ACCESSIBILITY = 1 << 9  # deprecated in PDF 2.0
    """(deprecated in PDF 2.0) Extract content for the purposes of accessibility.
    
    This bit should always be set for compatibility with processors supporting 
    earlier specifications.
    """

    ASSEMBLE_DOCUMENT = 1 << 10  # revision 3+, overrides MODIFY
    """For documents using revision 3 or later, assemble the document (i.e. insert, rotate, 
    and delete pages, create outlines, etc.), even if :attr:`.MODIFY` is clear."""

    FAITHFUL_PRINT = 1 << 11  # revision 3+
    """For documents using revision 3 or later, print the document in such a way that
    a faithful digital representation of the PDF can be generated. 
    
    If this bit is not set (and assuming :attr:`.PRINT` is also set), printing shall be 
    limited to a low-level representation, possibly of lower quality.
    """



from __future__ import annotations

import datetime
from io import BytesIO

import pdfnaut
from pdfnaut import PdfDocument
from pdfnaut.objects.xmp import XmpMetadata


def test_docinfo_read() -> None:
    pdf = PdfDocument.from_filename(r"tests\docs\sample.pdf")

    assert pdf.doc_info is not None

    assert pdf.doc_info.creator == "Rave (http://www.nevrona.com/rave)"

    assert pdf.doc_info.creation_date == datetime.datetime(
        2006, 3, 1, 7, 28, 26, tzinfo=datetime.timezone.utc
    )


def test_docinfo_write() -> None:
    original_pdf = PdfDocument.from_filename(r"tests\docs\sample.pdf")
    assert original_pdf.doc_info is not None

    # we don't need the microsecond since iso 8824 doesn't support them
    modify_date = datetime.datetime(2025, 3, 2, 15, 51, 0, tzinfo=datetime.timezone.utc)

    original_pdf.doc_info.title = "Sample PDF file"
    original_pdf.doc_info.modify_date = modify_date

    original_pdf.save((fp := BytesIO()))

    fp.seek(0)

    edited_pdf = PdfDocument(fp.read())
    assert edited_pdf.doc_info is not None

    assert edited_pdf.doc_info.title == "Sample PDF file"
    assert edited_pdf.doc_info.modify_date == modify_date


def test_docinfo_remove() -> None:
    original_pdf = PdfDocument.from_filename(r"tests\docs\sample.pdf")
    original_pdf.doc_info = None

    original_pdf.save((fp := BytesIO()))

    fp.seek(0)

    edited_pdf = PdfDocument(fp.read())
    assert edited_pdf.doc_info is None


def test_xmp_read() -> None:
    original_pdf = PdfDocument.from_filename(r"tests\docs\pdf2-incremental.pdf")

    assert original_pdf.xmp_info is not None

    # Text property
    assert original_pdf.xmp_info.pdf_producer == "Datalogics - example producer program name here"

    # Date property
    assert original_pdf.xmp_info.xmp_create_date == datetime.datetime(
        2017, 5, 24, 10, 30, 11, tzinfo=datetime.timezone.utc
    )

    # Language alternate property
    assert original_pdf.xmp_info.dc_title == {"x-default": "A simple PDF 2.0 example file"}

    # List property
    assert original_pdf.xmp_info.dc_creator == ["Datalogics Incorporated"]


def test_xmp_write() -> None:
    original_pdf = PdfDocument.from_filename(r"tests\docs\river-rle-image.pdf")

    original_pdf.xmp_info = XmpMetadata()

    # Text property
    original_pdf.xmp_info.pdf_producer = f"pdfnaut {pdfnaut.__version__}"

    # Date property
    dt = datetime.datetime.now(datetime.timezone.utc)

    original_pdf.xmp_info.xmp_create_date = dt
    original_pdf.xmp_info.xmp_modify_date = dt

    # Language alternate property
    original_pdf.xmp_info.dc_title = {"x-default": "The Tetons and the Snake River"}

    # List property
    original_pdf.xmp_info.dc_creator = ["Ansel Adams"]

    original_pdf.save((fp := BytesIO()))

    fp.seek(0)

    edited_pdf = PdfDocument(fp.read())
    assert edited_pdf.xmp_info is not None

    assert edited_pdf.xmp_info.pdf_producer == f"pdfnaut {pdfnaut.__version__}"
    assert edited_pdf.xmp_info.xmp_create_date == dt
    assert edited_pdf.xmp_info.xmp_modify_date == dt
    assert edited_pdf.xmp_info.dc_title == {"x-default": "The Tetons and the Snake River"}
    assert edited_pdf.xmp_info.dc_creator == ["Ansel Adams"]


def test_xmp_remove() -> None:
    original_pdf = PdfDocument.from_filename(r"tests\docs\pdf2-incremental.pdf")
    original_pdf.xmp_info = None

    original_pdf.save((fp := BytesIO()))

    fp.seek(0)

    edited_pdf = PdfDocument(fp.read())
    assert edited_pdf.xmp_info is None



from __future__ import annotations

import enum
from typing import TYPE_CHECKING, Generator, Literal, cast

from typing_extensions import Self

from ..common.fields import FlagField, NameField, StandardField, TextStringField
from ..cos.objects.base import PdfName, PdfReference
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.tokenizer import ContentStreamTokenizer

if TYPE_CHECKING:
    from ..cos.objects.stream import PdfStream


AnnotationKind = Literal[
    "Text",
    "Link",
    "FreeText",
    "Line",
    "Square",
    "Circle",
    "Polygon",
    "PolyLine",
    "Highlight",
    "Underline",
    "Squiggly",
    "StrikeOut",
    "Caret",
    "Stamp",
    "Ink",
    "Popup",
    "FileAttachment",
    "Sound",
    "Movie",
    "Screen",
    "Widget",
    "PrinterMark",
    "TrapNet",
    "Watermark",
    "3D",
    "Redact",
    "Projection",
    "RichMedia",
]


class AnnotationFlags(enum.IntFlag):
    """Flags for a particular annotation. See § 12.5.3, "Annotation flags" for details."""

    Null = 0
    """A default value."""

    Invisible = 1 << 0
    """For non-standard annotation types, do not render or print the annotation.
    If not set, the annotation shall be rendered according to its appearance stream.
    """

    Hidden = 1 << 1
    """Do not render the annotation or allow user interaction with it."""

    Print = 1 << 2
    """Print the annotation when the page is printed, except where the Hidden flag
    is set. If clear, do not print the annotation.
    """

    NoZoom = 1 << 3
    """Do not scale the annotation's appearance to the page's zoom factor."""

    NoRotate = 1 << 4
    """Do not rotate the annotation to match the page's rotation."""

    NoView = 1 << 5
    """Do not render the annotation or allow user interaction with it, but still
    allow printing according to the Print flag.
    """

    ReadOnly = 1 << 6
    """Do not allow user interaction with the annotation. This is ignored for Widget
    annotations."""

    Locked = 1 << 7
    """Do not allow the annotation to be removed or its properties to be modified
    but still allow its contents to be modified.
    """

    ToggleNoView = 1 << 8
    """Toggle the NoView flag when selecting or hovering over the annotation."""

    LockedContents = 1 << 9
    """Do not allow the contents of the annotation to be modified."""


class Annotation(PdfDictionary):
    """An annotation associates an object such as a note, link or rich media element
    with a location on a page of a PDF document (see § 12.5, "Annotations")."""

    kind = NameField[AnnotationKind]("Subtype")
    """The type of annotation. See "Table 171: Annotation types" for details."""

    rect = StandardField[PdfArray["int | float"]]("Rect")
    """A rectangle specifying the location of the annotation in the page."""

    contents = TextStringField("Contents")
    """The text contents that shall be displayed when the annotation is open, or if this
    annotation kind does not display text, an alternate description of the annotation's 
    contents."""

    name = TextStringField("NM")
    """An annotation name uniquely identifying it among other annotations in its page."""

    last_modified = TextStringField("M")
    """The date and time the annotation was most recently modified. This value should
    be a PDF date string but processors are expected to accept any text string."""

    flags = FlagField("F", AnnotationFlags, AnnotationFlags.Null)
    """A set of flags specifying various characteristics of the annotation."""

    color = StandardField["PdfArray[float] | None"]("C", None)
    """An array of 0 to 4 numbers in the range 0.0 to 1.0, representing a color used
    for the following purposes: 
        
    - The background of the annotation's icon when closed
    - The title bar of the annotation's popup window
    - The border of a link annotation.
    
    The number of array elements determines the color space in which the color shall
    be defined: 0 is no color, transparent; 1 is DeviceGray (grayscale); 3 is DeviceRGB;
    and 4 is DeviceCMYK 
    """

    language = TextStringField("Lang")
    """(PDF 2.0) A language identifier that shall specify the natural language for all 
    text in the annotation except where overridden by other explicit language 
    specifications (see § 14.9.2, "Natural language specification")."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls()
        dictionary.data = mapping.data

        return dictionary


class Page(PdfDictionary):
    """A page in the document (see § 7.7.3.3, "Page Objects").

    Arguments:
        size (tuple[int, int]):
            The width and height of the physical medium in which the page should
            be printed or displayed. Values provided in multiples of 1/72 of an inch.

        indirect_ref (PdfReference, optional):
            The indirect reference that this page object represents.
            In typical usage, this parameter should be none.
    """

    resources = StandardField["PdfDictionary | None"]("Resources", None)
    """Resources required by the page contents.

    If the page requires no resources, this returns an empty resource dictionary.
    If the page inherits its resources from an ancestor, this returns None.
    """

    mediabox = StandardField[PdfArray["int | float"]]("MediaBox")
    """A rectangle defining the boundaries of the physical medium in which the page
    should be printed or displayed."""

    cropbox = StandardField["PdfArray[int | float] | None"]("CropBox", None)
    """A rectangle defining the visible region of the page."""

    bleedbox = StandardField["PdfArray[int | float] | None"]("BleedBox", None)
    """A rectangle defining the region to which the contents of the page shall be 
    clipped when output in a production environment."""

    trimbox = StandardField["PdfArray[int | float] | None"]("TrimBox", None)
    """A rectangle defining the intended dimensions of the finished page after trimming."""

    artbox = StandardField["PdfArray[int | float] | None"]("ArtBox", None)
    """A rectangle defining the extent of the page's meaningful content as intended 
    by the page's creator."""

    user_unit = StandardField["int | float"]("UserUnit", 1)
    """The size of a user space unit, in multiples of 1/72 of an inch."""

    rotation = StandardField[int]("Rotate", 0)
    """The number of degrees by which the page shall be visually rotated clockwise.
    The value is a multiple of 90 (by default, 0)."""

    metadata = StandardField["PdfStream | None"]("Metadata", None)
    """A metadata stream, generally written in XMP, containing information about this page."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary, indirect_ref: PdfReference | None = None) -> Self:
        dictionary = cls(size=(0, 0), indirect_ref=indirect_ref)
        dictionary.data = mapping.data

        return dictionary

    def __init__(
        self, size: tuple[int | float, int | float], *, indirect_ref: PdfReference | None = None
    ) -> None:
        super().__init__()

        self.indirect_ref = indirect_ref

        self["Type"] = PdfName(b"Page")
        self["MediaBox"] = PdfArray([0, 0, *size])

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} mediabox={self.mediabox!r} rotation={self.rotation!r}>"

    @property
    def content_stream(self) -> ContentStreamTokenizer | None:
        """An iterator over the instructions producing the contents of this page."""
        if "Contents" not in self:
            return

        contents = cast("PdfStream | PdfArray[PdfStream]", self["Contents"])

        if isinstance(contents, PdfArray):
            # in case the Contents of a document are an array, they must be
            # concatenated into a single one.
            return ContentStreamTokenizer(b"\n".join(stm.decode() for stm in contents))

        return ContentStreamTokenizer(contents.decode())

    @property
    def annotations(self) -> Generator[Annotation, None, None]:
        """All annotations associated with this page (see § 12.5, "Annotations" and :class:`.Annotation`)."""
        for annot in cast(PdfArray[PdfDictionary], self.get("Annots", PdfArray())):
            yield Annotation.from_dict(annot)



from __future__ import annotations

from .catalog import PageLayout, PageMode, UserAccessPermissions
from .page import Annotation, AnnotationFlags, Page
from .trailer import Info, TrappedState
from .xmp import XmpMetadata

__all__ = (
    "PageLayout",
    "PageMode",
    "Page",
    "Annotation",
    "AnnotationFlags",
    "Info",
    "TrappedState",
    "UserAccessPermissions",
    "XmpMetadata",
)



"""Utilities for parsing and encoding date formats: ISO 8601 and ISO 8824."""

from __future__ import annotations

import datetime
import re


def has_date(date: datetime.datetime) -> bool:
    """Returns whether ``date`` has a date component. In this case, if either the year,
    month or day isn't a default value."""
    return date.year > 1 or date.month > 1 or date.day > 1


def has_time(date: datetime.datetime) -> bool:
    """Returns whether ``date`` has a time component. In this case, if either the hour,
    minute, second or microsecond isn't a default value."""
    return date.hour > 0 or date.minute > 0 or date.second > 0 or date.microsecond > 0


def has_timezone(date: datetime.datetime) -> bool:
    """Returns whether ``date`` specifies a timezone other than UTC."""
    offset = date.utcoffset()
    return offset is not None and offset.total_seconds() != 0


def parse_iso8824(date_string: str) -> datetime.datetime:
    """Parses an ISO/IEC 8824 date string into a :class:`datetime.datetime` object
    (for example, ``D:20010727133720``).

    This is the type of date string described in § 7.9.4, "Dates" of the PDF spec.
    """

    # dates may end with an apostrophe (pdf 1.7 and below)
    if date_string.endswith("'"):
        date_string = date_string[:-1]

    pattern = re.compile(
        r"""^D:(?P<year>\d{4})(?P<month>\d{2})?(?P<day>\d{2})?                       # date
                (?P<hour>\d{2})?(?P<minute>\d{2})?(?P<second>\d{2})?                 # time
                (?P<offset>[-+Z])?(?P<offset_hour>\d{2})?(?P<offset_minute>'\d{2})?$ # offset
        """,
        re.X,
    )

    mat = pattern.match(date_string)
    if not mat:
        raise ValueError(f"Invalid date format: {date_string!r}")

    offset_sign = mat.group("offset")
    if offset_sign is None or offset_sign == "Z":
        offset_hour = 0
        offset_minute = 0
    else:
        offset_hour = int(mat.group("offset_hour") or 0)
        offset_minute = int((mat.group("offset_minute") or "'0")[1:])

    if offset_sign == "-":
        offset_hour = -offset_hour

    delta = datetime.timedelta(hours=offset_hour, minutes=offset_minute)

    return datetime.datetime(
        year=int(mat.group("year")),
        month=int(mat.group("month") or 1),
        day=int(mat.group("day") or 1),
        hour=int(mat.group("hour") or 0),
        minute=int(mat.group("minute") or 0),
        second=int(mat.group("second") or 0),
        tzinfo=datetime.timezone(delta),
    )


def encode_iso8824(date: datetime.datetime, *, full: bool = True) -> str:
    """Encodes a :class:`datetime.datetime` object into an ISO 8824 date string suitable
    for storage in a PDF file.

    If ``full`` is True, this function will encode all date and time values. Otherwise,
    the function will perform partial encoding, only including components that aren't
    their default values.
    """

    datestr = f"D:{date.year}"

    if has_date(date) or full:
        datestr += f"{date.month:02}" if date.month > 1 or date.day > 1 or full else ""
        datestr += f"{date.day:02}" if date.day > 1 or date.second > 0 or full else ""

    if has_time(date) or has_timezone(date) or full:
        datestr += f"{date.hour:02}" if date.hour > 0 or date.minute > 0 or full else ""
        datestr += f"{date.minute:02}" if date.minute > 0 or date.second > 0 or full else ""
        datestr += f"{date.second:02}" if date.second > 0 or has_timezone(date) or full else ""

        offset = date.utcoffset()
        # If no offset, assume UTC
        if offset is None or offset.total_seconds() == 0:
            return datestr + "Z"

        offset_hours, offset_seconds = divmod(offset.total_seconds(), 3600)
        offset_minutes = int(offset_seconds / 60)

        return datestr + f"{offset_hours:+03n}'{offset_minutes:02}"

    return datestr


def parse_iso8601(date_string: str) -> datetime.datetime:
    """Parses a date string conforming to the ISO 8601 profile specified in
    https://www.w3.org/TR/NOTE-datetime into a :class:`datetime.datetime` object."""

    pattern = re.compile(
        r"""^(?P<year>\d{4})(?:-(?P<month>\d{2}))?(?:-(?P<day>\d{2}))? # yyyy-mm-dd
             (?:T(?P<hour>\d{2}):(?P<minute>\d{2})                     # hh-mm
             (?::(?P<second>\d{2})(?:\.(?P<fraction>\d+))?)?           # ss.s
             (?P<tzd>Z|[-+]\d{2}:\d{2})?)?$                            # Z or +hh:mm
        """,
        re.X,
    )

    mat = pattern.match(date_string)
    if not mat:
        raise ValueError(f"Expected an ISO 8601 string, received {date_string!r}")

    tzd = mat.group("tzd")
    if tzd is None or tzd == "Z":
        tz_offset = datetime.timedelta(hours=0, minutes=0)
    else:
        hh, mm = tzd.split(":")
        tz_offset = datetime.timedelta(hours=int(hh), minutes=int(mm))

    fraction_str = mat.group("fraction") or "0"
    fraction_micro = (int(fraction_str) / 10 ** len(fraction_str)) * 1_000_000

    return datetime.datetime(
        year=int(mat.group("year")),
        month=int(mat.group("month") or 1),
        day=int(mat.group("day") or 1),
        hour=int(mat.group("hour") or 0),
        minute=int(mat.group("minute") or 0),
        second=int(mat.group("second") or 0),
        microsecond=int(fraction_micro),
        tzinfo=datetime.timezone(tz_offset),
    )


def encode_iso8601(date: datetime.datetime, *, full: bool = True) -> str:
    """Encodes a :class:`datetime.datetime` object into a date string conforming to the
    ISO 6801 profile specified in https://www.w3.org/TR/NOTE-datetime.

    If ``full`` is True, this function will encode all date and time values. Otherwise,
    the function will perform partial encoding, only including components that aren't
    their default values.
    """

    datestr = str(date.year)

    if has_date(date) or full:
        datestr += f"-{date.month:02}"
        # Append day if present or if hour present
        datestr += f"-{date.day:02}" if date.day > 1 or date.hour > 0 or full else ""

    if has_time(date) or has_timezone(date) or full:
        datestr += f"T{date.hour:02}:{date.minute:02}"

        # Whether we have a second
        if date.second > 0 or date.microsecond > 0 or full:
            datestr += f":{date.second:02}"
            datestr += f".{date.microsecond:06n}" if date.microsecond > 0 else ""

        offset = date.utcoffset()
        # If no offset, assume UTC
        if offset is None or offset.total_seconds() == 0:
            return datestr + "Z"

        offset_hours, offset_seconds = divmod(offset.total_seconds(), 3600)
        offset_minutes = int(offset_seconds / 60)

        return datestr + f"{offset_hours:+03n}:{offset_minutes:02}"

    return datestr



# Unit tests for tokenizing the COS syntax in PDFs.

from __future__ import annotations

from typing import cast

from pdfnaut.cos.objects import (
    PdfArray,
    PdfComment,
    PdfDictionary,
    PdfHexString,
    PdfName,
    PdfNull,
    PdfReference,
)
from pdfnaut.cos.objects.base import PdfOperator
from pdfnaut.cos.tokenizer import ContentStreamTokenizer, PdfTokenizer


def test_null_and_boolean() -> None:
    lexer = PdfTokenizer(b"null true false")
    tokens = list(lexer)

    assert isinstance(tokens[0], PdfNull)
    assert tokens[1] is True and tokens[2] is False


def test_numeric() -> None:
    lexer = PdfTokenizer(b"-1 +25 46 -32.591 +52.871 3.1451 -.302 3. .25")
    tokens = list(lexer)

    assert tokens == [-1, 25, 46, -32.591, 52.871, 3.1451, -0.302, 3.0, 0.25]


def test_name_object() -> None:
    lexer = PdfTokenizer(b"/Type /SomeR@ndomK*y /Lime#20Green / /F#23")
    tokens = list(lexer)
    assert tokens == [
        PdfName(b"Type"),
        PdfName(b"SomeR@ndomK*y"),
        PdfName(b"Lime Green"),
        PdfName(b""),
        PdfName(b"F#"),
    ]


def test_literal_string() -> None:
    # Basic string
    lexer = PdfTokenizer(b"(The quick brown fox jumps over the lazy dog.)")
    assert lexer.get_next_token() == b"The quick brown fox jumps over the lazy dog."

    # String with nested parentheses
    lexer = PdfTokenizer(b"(This is a string with a (few) nested ((parentheses)))")
    assert lexer.get_next_token() == b"This is a string with a (few) nested ((parentheses))"

    # String continued in next line
    lexer = PdfTokenizer(b"(This is a string that is \r\ncontinued on the next line)")
    assert lexer.get_next_token() == b"This is a string that is \r\ncontinued on the next line"

    # String ending with a \ at the EOL and followed next line
    lexer = PdfTokenizer(b"(This is a string \\\r\nwith no newlines.)")
    assert lexer.get_next_token() == b"This is a string with no newlines."

    # String with escape characters
    lexer = PdfTokenizer(b"(This is a string with a \\t tab character and a \\053 plus.))")
    assert lexer.get_next_token() == b"This is a string with a \t tab character and a + plus."

    # String with contiguous octal sequences
    lexer = PdfTokenizer(b"(\\110\\151\\41)")
    assert lexer.get_next_token() == b"Hi!"

    # String with invalid octal sequences
    lexer = PdfTokenizer(b"(\\318 then \\387 and then \\981)")
    assert lexer.get_next_token() == b"\318 then \387 and then 981"


def test_hex_string() -> None:
    lexer = PdfTokenizer(b"<A5B2FF><6868ADE>")
    tokens = cast("list[PdfHexString]", list(lexer))

    assert tokens[0].raw == b"A5B2FF" and tokens[1].raw == b"6868ADE0"


def test_dictionary() -> None:
    lexer = PdfTokenizer(b"""<< /Type /Catalog /Metadata 2 0 R /Pages 3 0 R >>""")
    assert lexer.get_next_token() == PdfDictionary(
        {"Type": PdfName(b"Catalog"), "Metadata": PdfReference(2, 0), "Pages": PdfReference(3, 0)}
    )


def test_comment_and_eol() -> None:
    lexer = PdfTokenizer(b"% This is a comment\r\n12 % This is another comment\r25\n")
    assert isinstance(com := next(lexer), PdfComment) and com.value == b" This is a comment"
    assert next(lexer) == 12
    assert isinstance(com := next(lexer), PdfComment) and com.value == b" This is another comment"
    assert next(lexer) == 25

    lexer = PdfTokenizer(b"% This is a comment ending with \\r\r")
    assert (
        isinstance(com := lexer.get_next_token(), PdfComment)
        and com.value == b" This is a comment ending with \\r"
    )


def test_array() -> None:
    # Simple array
    lexer = PdfTokenizer(b"[45 <</Size 40>> (42)]")
    assert lexer.get_next_token() == PdfArray([45, {"Size": 40}, b"42"])

    # Nested array
    lexer = PdfTokenizer(b"[/XYZ [45 32 76] /Great]")
    assert lexer.get_next_token() == PdfArray([PdfName(b"XYZ"), [45, 32, 76], PdfName(b"Great")])


def test_indirect_reference() -> None:
    lexer = PdfTokenizer(b"2 0 R")
    assert lexer.get_next_token() == PdfReference(2, 0)


def test_content_stream() -> None:
    # test that we aren't parsing the references
    lexer = ContentStreamTokenizer(b"1 0 0 RG 0 w")
    assert list(lexer) == [PdfOperator(b"RG", [1, 0, 0]), PdfOperator(b"w", [0])]

    # test without padding
    lexer = ContentStreamTokenizer(b"/F29 12 Tf 100 740 Td[(Lorem)-447(ipsum)-446(text.)]TJ")

    assert list(lexer) == [
        PdfOperator(b"Tf", [PdfName(b"F29"), 12]),
        PdfOperator(b"Td", [100, 740]),
        PdfOperator(b"TJ", [PdfArray([b"Lorem", -447, b"ipsum", -446, b"text."])]),
    ]

    # test that we are parsing other objects
    lexer = ContentStreamTokenizer(b"/F1 12 Tf 72 712 Td (A stream with an indirect length) Tj")
    assert list(lexer) == [
        PdfOperator(b"Tf", [PdfName(b"F1"), 12]),
        PdfOperator(b"Td", [72, 712]),
        PdfOperator(b"Tj", [b"A stream with an indirect length"]),
    ]

    # test that we are parsing an inline image
    lexer = ContentStreamTokenizer(b"""\
BI              % Begin inline image object
    /W 17 /H 17 % Width and height in samples
    /CS /RGB    % Color space
    /BPC 8      % Bits per component
    /F [/A85]   % Filters
ID              % Begin image data
E,8rmAS?!uA7]d(AoD]4Bl.9kAH~>
EI              % End inline image object
    """)

    operator = list(lexer)

    assert operator[0].name == b"EI"
    assert operator[0].args[0].details == PdfDictionary(
        {
            "W": 17,
            "H": 17,
            "CS": PdfName(b"RGB"),
            "BPC": 8,
            "F": PdfArray([PdfName(b"A85")]),
        }
    )
    assert operator[0].args[0].raw == b"E,8rmAS?!uA7]d(AoD]4Bl.9kAH~>\n"



from __future__ import annotations

from typing import Generator, cast

from .cos.objects import (
    PdfArray,
    PdfDictionary,
    PdfHexString,
    PdfName,
    PdfReference,
    PdfStream,
)
from .cos.objects.base import parse_text_string
from .cos.objects.xref import FreeXRefEntry, InUseXRefEntry, PdfXRefEntry
from .cos.parser import FreeObject, PdfParser, PermsAcquired
from .cos.serializer import PdfSerializer
from .objects.catalog import PageLayout, PageMode, UserAccessPermissions
from .objects.page import Page
from .objects.trailer import Info
from .objects.xmp import XmpMetadata
from .page_list import PageList, flatten_pages


class PdfDocument(PdfParser):
    """A high-level interface over :class:`~.PdfParser`.

    PDF authors who want to work with a document in a high-level way should
    use this interface over ``PdfParser``.
    """

    @classmethod
    def from_filename(cls, path: str, *, strict: bool = False) -> PdfDocument:
        """Loads a PDF document from a file ``path``."""
        with open(path, "rb") as fp:
            return PdfDocument(fp.read(), strict=strict)

    @classmethod
    def new(cls) -> PdfDocument:
        """Creates a blank PDF document."""

        builder = PdfSerializer()
        builder.write_header("2.0")

        builder.objects[(1, 0)] = PdfDictionary(
            {"Type": PdfName(b"Catalog"), "Pages": PdfReference(2, 0)}
        )
        builder.objects[(2, 0)] = PdfDictionary(
            {"Type": PdfName(b"Pages"), "Kids": PdfArray(), "Count": 0}
        )

        section: list[tuple[int, PdfXRefEntry]] = [(0, FreeXRefEntry(0, 65535))]

        for (obj_num, gen_num), item in builder.objects.items():
            offset = builder.write_object((obj_num, gen_num), item)
            section.append((obj_num, InUseXRefEntry(offset, gen_num)))

        subsections = builder.generate_xref_section(section)

        startxref = builder.write_standard_xref_section(subsections)

        builder.write_trailer(
            PdfDictionary({"Size": subsections[0].count, "Root": PdfReference(1, 0)}), startxref
        )

        builder.write_eof()

        return PdfDocument(builder.content)

    def __init__(self, data: bytes, *, strict: bool = False) -> None:
        super().__init__(data, strict=strict)

        self.parse()

        self.access_level = PermsAcquired.OWNER
        """The current access level of the document, specified as a value from the
        :class:`.PermsAcquired` enum.

        - Owner (2): Full access to the document. If the document is not encrypted, \
        this is the default value.
        - User (1): Access to the document under restrictions.
        - None (0): Document is currently encrypted.
        """

        # files under permissions usually use an empty string as a password
        if self.has_encryption:
            self.access_level = self.decrypt("")

        self._page_list: PageList | None = None

    @property
    def has_encryption(self) -> bool:
        """Whether this document includes encryption."""
        return "Encrypt" in self.trailer

    @property
    def catalog(self) -> PdfDictionary:
        """The root of the document's object hierarchy, including references to pages,
        outlines, destinations, and other core elements of a PDF document.

        For details on the contents of the catalog, see § 7.7.2, "Document Catalog".
        """
        return cast(PdfDictionary, self.trailer["Root"])

    @catalog.setter
    def catalog(self, value: PdfDictionary) -> None:
        root_ref = cast(PdfReference, self.trailer.data["Root"])
        self.objects[root_ref.object_number] = value

    @property
    def doc_info(self) -> Info | None:
        """The ``/Info`` entry in the catalog which includes document-level information
        described in § 14.3.3, "Document information dictionary".

        Some documents may specify a metadata stream rather than a DocInfo dictionary.
        Such metadata can be accessed using :attr:`.PdfDocument.xmp_info`.

        PDF 2.0 deprecates all keys of the DocInfo dictionary except for ``CreationDate``
        and ``ModDate``.
        """
        if "Info" not in self.trailer:
            return

        return Info.from_dict(cast(PdfDictionary, self.trailer["Info"]))

    @doc_info.setter
    def doc_info(self, value: Info | None) -> None:
        info_ref = cast("PdfReference | None", self.trailer.data.get("Info"))

        # A new docinfo object will be created
        if info_ref is None and value is not None:
            new_object = max(self.objects) + 1
            self.objects[new_object] = PdfDictionary(**value.data)
            self.trailer.data["Info"] = PdfReference(new_object, 0).with_resolver(self.get_object)
        # A docinfo object will be set
        elif info_ref and isinstance(value, Info):
            self.objects[info_ref.object_number] = PdfDictionary(**value.data)
        # A docinfo object will be removed
        elif info_ref:
            self.objects[info_ref.object_number] = FreeObject()
            self.trailer.data.pop("Info", None)

    @property
    def pdf_version(self) -> str:
        """The version of the PDF standard implemented by this document.

        For obtaining the PDF version, the ``/Version`` entry in the catalog
        is checked. If no such key is present, the version specified in the
        header is returned. If both are present, the version returned is the
        latest specified according to lexicographical comparison.
        """
        header_version = self.header_version
        catalog_version = cast("PdfName | None", self.catalog.get("Version"))

        if not catalog_version:
            return header_version

        return max((header_version, catalog_version.value.decode()))

    @property
    def xmp_info(self) -> XmpMetadata | None:
        """The ``/Metadata`` entry of the catalog which includes document-level
        metadata stored as XMP."""
        if "Metadata" not in self.catalog:
            return

        stm = cast(PdfStream, self.catalog["Metadata"])
        return XmpMetadata(stm)

    @xmp_info.setter
    def xmp_info(self, xmp: XmpMetadata | None) -> None:
        metadata_ref = cast("PdfReference | None", self.catalog.data.get("Metadata"))

        # A new metadata object will be created
        if metadata_ref is None and xmp is not None:
            self.catalog["Metadata"] = self.objects.add(xmp.stream)
        # A metadata object will be set
        elif metadata_ref and isinstance(xmp, XmpMetadata):
            self.objects[metadata_ref.object_number] = xmp.stream
        # A metadata object will be removed
        elif metadata_ref:
            self.objects.delete(metadata_ref.object_number)
            self.catalog.pop("Metadata", None)

    @property
    def page_tree(self) -> PdfDictionary:
        """The document's page tree. See § 7.7.3, "Page Tree" for details.

        For iterating over the pages of a PDF, prefer :attr:`PdfDocument.pages`
        or :attr:`.PdfDocument.flattened_pages`.
        """
        return cast(PdfDictionary, self.catalog["Pages"])

    @property
    def outline_tree(self) -> PdfDictionary | None:
        """The document's outline tree including what is commonly referred to as
        bookmarks. See § 12.3.3, "Document Outline" for details.
        """
        return cast("PdfDictionary | None", self.catalog.get("Outlines"))

    def decrypt(self, password: str) -> PermsAcquired:
        self.access_level = super().decrypt(password)
        return self.access_level

    @property
    def flattened_pages(self) -> Generator[Page, None, None]:
        """A generator suitable for iterating over the pages of a PDF."""
        return flatten_pages(self.page_tree)

    @property
    def page_layout(self) -> PageLayout:
        """The page layout to use when opening the document. May be one of the following
        values:

        - SinglePage: Display one page at a time (default).
        - OneColumn: Display the pages in one column.
        - TwoColumnLeft: Display the pages in two columns, with odd-numbered pages
          on the left.
        - TwoColumnRight: Display the pages in two columns, with odd-numbered pages
          on the right.
        - TwoPageLeft: Display the pages two at a time, with odd-numbered
          pages on the left (PDF 1.5).
        - TwoPageRight: Display the pages two at a time, with odd-numbered
          pages on the right (PDF 1.5).
        """
        if "PageLayout" not in self.catalog:
            return "SinglePage"

        return cast(PageLayout, cast(PdfName, self.catalog["PageLayout"]).value.decode())

    @property
    def page_mode(self) -> PageMode:
        """Value specifying how the document shall be displayed when opened:

        - UseNone: Neither document outline nor thumbnail images visible (default).
        - UseOutlines: Document outline visible.
        - UseThumbs: Thumbnail images visible.
        - FullScreen: Full-screen mode, with no menu bar, window controls, or any
          other window visible.
        - UseOC: Optional content group panel visible (PDF 1.5).
        - UseAttachments: Attachments panel visible (PDF 1.6).
        """
        if "PageMode" not in self.catalog:
            return "UseNone"

        return cast(PageMode, cast(PdfName, self.catalog["PageMode"]).value.decode())

    @property
    def language(self) -> str | None:
        """A language identifier that shall specify the natural language for all text in
        the document except where overridden by language specifications for structure
        elements or marked content (see § 14.9.2, "Natural language specification").

        If this entry is absent, the language shall be considered unknown.
        """

        if "Lang" not in self.catalog:
            return

        return parse_text_string(cast("PdfHexString | bytes", self.catalog["Lang"]))

    @property
    def access_permissions(self) -> UserAccessPermissions | None:
        """User access permissions relating to the document if any.

        See "Table 22: User Access Permissions" and :class:`.UserAccessPermissions`.
        """
        if not self.has_encryption:
            return

        encrypt_dict = cast(PdfDictionary, self.trailer["Encrypt"])

        if (perms := encrypt_dict.get("P")) is not None:
            return UserAccessPermissions(perms)

    @property
    def pages(self) -> PageList:
        """The page list in the document."""

        if not self.access_level:
            raise PermissionError("Cannot read pages of encrypted document.")

        if self._page_list is None:
            self._page_list = PageList(
                self, self.page_tree, cast(PdfReference, self.catalog.data["Pages"])
            )

        return self._page_list



from __future__ import annotations

import datetime
import xml.dom.minidom as minidom
from textwrap import dedent
from typing import Any
from xml.parsers import expat

import pdfnaut
from pdfnaut.common.dates import encode_iso8601, parse_iso8601
from pdfnaut.cos.objects import PdfDictionary, PdfName, PdfStream
from pdfnaut.exceptions import PdfParseError
from pdfnaut.objects.trailer import TrappedState

namespaces = {
    "pdf": "http://ns.adobe.com/pdf/1.3/",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "dc": "http://purl.org/dc/elements/1.1/",
    "xmp": "http://ns.adobe.com/xap/1.0/",
}


def get_full_text(element: minidom.Element) -> str:
    """Returns the full text content within ``element``."""

    text_values = []

    for node in element.childNodes:
        if isinstance(node, minidom.Text):
            text_values.append(node.data)
        elif hasattr(node, "childNodes"):
            text_values.append(get_full_text(node))

    return "".join(text_values)


def lookup_prefix_for_ns(node: minidom.Node, namespace: str) -> tuple[str, minidom.Node] | None:
    """Locates a namespace prefix matching the ``namespace`` URI. Returns either
    a tuple of two items containing, in order, the prefix of the namespace URI and the
    node where it was found, or None, if no prefix is registered for the namespace URI.

    This is an implementation of https://dom.spec.whatwg.org/#locate-a-namespace-prefix.
    """

    if isinstance(node, minidom.Element):
        if node.namespaceURI == namespace and node.prefix:
            return (node.prefix, node)

        for attrib in node.attributes.values():
            if attrib.prefix == "xmlns" and attrib.value == namespace:
                return (attrib.localName, node)

        if node.parentNode:
            return lookup_prefix_for_ns(node.parentNode, namespace)
    elif isinstance(node, minidom.Document):
        if node.ownerDocument is None:
            return

        return lookup_prefix_for_ns(node.ownerDocument, namespace)
    elif isinstance(node, (minidom.DocumentFragment, minidom.DocumentType)):
        return
    elif isinstance(node, minidom.Attr):
        if node.ownerElement is None:
            return

        return lookup_prefix_for_ns(node.ownerElement, namespace)
    elif node.parentNode is not None:
        return lookup_prefix_for_ns(node.parentNode, namespace)


class XMPProperty:
    """An XMP property included in an XMP packet."""

    def __init__(self, namespace_uri: str, local_name: str, **extra: Any) -> None:
        self.namespace_uri = namespace_uri
        """The namespace URI of this property."""

        self.local_name = local_name
        """The local name of this property."""

        self.extra = extra
        """Any additional property-specific values."""

        self._xml_property: minidom.Element | None = None

    def _fetch_xml_property(self, xmp: XmpMetadata) -> None:
        if self._xml_property is not None:
            return

        candidates = xmp.rdf_root.getElementsByTagNameNS(self.namespace_uri, self.local_name)
        self._xml_property = candidates[0] if candidates else None

    def _set_xml_property(self, xmp: XmpMetadata, node_list: list[minidom.Node]) -> None:
        self._fetch_xml_property(xmp)

        if self._xml_property:
            # This property is present in the document.
            # Simply replace the children with a new text node.
            self._xml_property.childNodes[:] = node_list
            xmp.stream.modify(xmp.packet.toprettyxml().encode())

            return

        # We will have to make a new property
        # Let's first check the document to see if our namespace is already registered.
        prefix_and_node = lookup_prefix_for_ns(xmp.rdf_root, self.namespace_uri)
        if prefix_and_node is not None:
            prefix, element_with_prefix = prefix_and_node
        else:
            prefix, element_with_prefix = None, None

        # Then check the children
        for child in xmp.rdf_root.childNodes:
            if prefix_and_node := lookup_prefix_for_ns(child, self.namespace_uri):
                prefix, element_with_prefix = prefix_and_node
                break

        if prefix and element_with_prefix:
            # An element with this namespace is registered, create the element.
            element = xmp.packet.createElementNS(self.namespace_uri, f"{prefix}:{self.local_name}")
        else:
            # No prefix for namespace in document, register it in parent.

            prefix_by_ns = {prefix: ns for ns, prefix in namespaces.items()}
            xmp.rdf_root.setAttribute(
                f"xmlns:{prefix_by_ns[self.namespace_uri]}", self.namespace_uri
            )

            element = xmp.packet.createElementNS(
                self.namespace_uri, prefix_by_ns[self.namespace_uri] + ":" + self.local_name
            )

            element_with_prefix = xmp.rdf_root

        # Insert the new element
        element.childNodes[:] = node_list
        element_with_prefix.appendChild(element)

        xmp.stream.modify(xmp.packet.toprettyxml().encode())

    def _ensure_rdf_prefix(self, xmp: XmpMetadata) -> str:
        prefix_and_node = lookup_prefix_for_ns(xmp.rdf_root, namespaces["rdf"])

        if prefix_and_node is not None:
            prefix, _ = prefix_and_node
        else:
            xmp.rdf_root.setAttribute("xmlns:rdf", namespaces["rdf"])
            prefix = "rdf"

        return prefix

    def _delete_xml_property(self, xmp: XmpMetadata) -> None:
        self._fetch_xml_property(xmp)

        if self._xml_property is None:
            raise ValueError(
                f"No such element: {self.local_name!r} with namespace {self.namespace_uri!r}"
            )

        self._xml_property.parentNode.removeChild(self._xml_property)
        self._xml_property = None


class XMPTextProperty(XMPProperty):
    """An XMP Text property -- a possibly empty Unicode string."""

    def __get__(self, xmp: XmpMetadata, objtype: Any | None = None) -> str | None:
        self._fetch_xml_property(xmp)

        return get_full_text(self._xml_property) if self._xml_property else None

    def __set__(self, xmp: XmpMetadata, value: str) -> None:
        self._fetch_xml_property(xmp)

        text_node = xmp.packet.createTextNode(value)
        self._set_xml_property(xmp, [text_node])

    def __delete__(self, xmp: XmpMetadata) -> None:
        self._delete_xml_property(xmp)


class XMPLangAltProperty(XMPProperty):
    """An XMP Language Alternative property -- an alternative array of simple text items
    facilitating the selection of a text item based on a desired language.

    In this case, this array is represented as a mapping of language names to text items
    corresponding to each language. The language name should be a value as defined in RFC
    3066, composed of a primary language subtag and an optional series of subsequent subtags.

    The default value, if known, should be the first item in the dictionary. A default
    value may also be explicitly marked by setting its language to 'x-default'.

    See https://developer.adobe.com/xmp/docs/XMPNamespaces/XMPDataTypes/#language-alternative.
    """

    def __get__(self, xmp: XmpMetadata, objtype: Any | None) -> dict[str, str] | None:
        self._fetch_xml_property(xmp)

        if self._xml_property is None:
            return

        alt = self._xml_property.getElementsByTagNameNS(namespaces["rdf"], "Alt")
        if not alt:
            return

        langalt = {}

        for element in alt[0].getElementsByTagNameNS(namespaces["rdf"], "li"):
            langalt[element.attributes["xml:lang"].value] = get_full_text(element)

        return langalt

    def __set__(self, xmp: XmpMetadata, value: dict[str, str]) -> None:
        self._fetch_xml_property(xmp)

        prefix = self._ensure_rdf_prefix(xmp)
        alt: minidom.Element = xmp.packet.createElementNS(namespaces["rdf"], f"{prefix}:Alt")

        for lang, val in value.items():
            list_item: minidom.Element = xmp.packet.createElementNS(
                namespaces["rdf"], f"{prefix}:li"
            )
            list_item.setAttribute("xml:lang", lang)
            list_item.childNodes.append(xmp.packet.createTextNode(val))

            alt.appendChild(list_item)

        self._set_xml_property(xmp, [alt])

    def __delete__(self, xmp: XmpMetadata) -> None:
        self._delete_xml_property(xmp)


class XMPListProperty(XMPProperty):  # list being either a sequence or bag
    """An array valued XMP property -- in this context, either an RDF sequence, used
    for ordered arrays, or an RDF bag, used for unordered arrays.

    See § 7.7 "Array valued XMP properties" in Part 1 of the XMP specification.
    """

    def __get__(self, xmp: XmpMetadata, objtype: Any | None) -> list[Any] | None:
        self._fetch_xml_property(xmp)

        if self._xml_property is None:
            return

        containers = self._xml_property.getElementsByTagNameNS(
            namespaces["rdf"], self.extra["kind"]
        )
        if not containers:
            return

        items = []
        for element in containers[0].getElementsByTagNameNS(namespaces["rdf"], "li"):
            items.append(get_full_text(element))

        return items

    def __set__(self, xmp: XmpMetadata, value: list[Any]) -> None:
        self._fetch_xml_property(xmp)

        prefix = self._ensure_rdf_prefix(xmp)
        kind = self.extra["kind"]
        container: minidom.Element = xmp.packet.createElementNS(
            namespaces["rdf"], f"{prefix}:{kind}"
        )

        for item in value:
            list_item: minidom.Element = xmp.packet.createElementNS(
                namespaces["rdf"], f"{prefix}:li"
            )

            list_item.childNodes.append(xmp.packet.createTextNode(item))

            container.appendChild(list_item)

        self._set_xml_property(xmp, [container])

    def __delete__(self, xmp: XmpMetadata) -> None:
        self._delete_xml_property(xmp)


class XMPDateProperty(XMPProperty):
    """An XMP Date property -- an ISO 8601 date string, or specifically, the subset
    specified in https://www.w3.org/TR/NOTE-datetime.

    See https://developer.adobe.com/xmp/docs/XMPNamespaces/XMPDataTypes/#date.
    """

    def __get__(self, xmp: XmpMetadata, objtype: Any | None = None) -> datetime.datetime | None:
        self._fetch_xml_property(xmp)

        if self._xml_property is None:
            return

        text = get_full_text(self._xml_property)
        return parse_iso8601(text)

    def __set__(self, xmp: XmpMetadata, value: datetime.datetime) -> None:
        self._fetch_xml_property(xmp)

        text_node = xmp.packet.createTextNode(encode_iso8601(value))
        self._set_xml_property(xmp, [text_node])

    def __delete__(self, xmp: XmpMetadata) -> None:
        self._delete_xml_property(xmp)


class XmpMetadata:
    """An object representing Extensible Metadata Platform (XMP) metadata, either pertaining
    to an entire document or to a particular resource.

    For information about XMP, see https://developer.adobe.com/xmp/docs/.

    Arguments:
        stream (PdfStream, optional):
            The XMP packet to parse as a PDF stream. If ``stream`` is None,
            a new stream containing a packet will be created.

    Raises:
        PdfParseError: If ``stream`` does not contain a valid XMP packet.
    """

    def __init__(self, stream: PdfStream | None = None) -> None:
        self.stream: PdfStream
        """The XMP packet as a string."""

        if stream is None:
            self.stream = PdfStream.create(
                dedent(f"""\
                    <?xml version="1.0" ?>
                    <?xpacket begin="" id="W5M0MpCehiHzreSzNTczkc9d"?>
                    <x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="pdfnaut {pdfnaut.__version__}">
                        <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
                        </rdf:RDF>
                    </x:xmpmeta>
                    <?xpacket end="w"?>
                """).encode(),
                PdfDictionary(Type=PdfName(b"Metadata"), Subtype=PdfName(b"XML")),
            )
        else:
            self.stream = stream

        try:
            self.packet = minidom.parseString(self.stream.decode())
            """The XMP packet as an XML document."""
        except expat.ExpatError as exc:
            raise PdfParseError("Metadata value is not a valid XMP packet.") from exc

        self.rdf_root = self.packet.getElementsByTagNameNS(namespaces["rdf"], "RDF")[0]
        """The RDF root of the packet being parsed."""

    # * PDF namespace properties
    # * https://developer.adobe.com/xmp/docs/XMPNamespaces/pdf/
    # Note: I have also seen other properties for the PDF namespace,
    # such as Subject, Author, and Copyright. But, as far as I know,
    # they're not official and seem to stem from improper reconciling
    # of the DocInfo dictionary.

    pdf_producer = XMPTextProperty(namespaces["pdf"], "Producer")
    """The name of the tool that produced this PDF document."""

    pdf_keywords = XMPTextProperty(namespaces["pdf"], "Keywords")
    """Keywords associated with the document."""

    pdf_pdfversion = XMPTextProperty(namespaces["pdf"], "PDFVersion")
    """The PDF file version. For example, '1.0' or '1.3'."""

    _pdf_trapped_raw = XMPTextProperty(namespaces["pdf"], "Trapped")

    @property
    def pdf_trapped(self) -> TrappedState:
        """A value indicating whether the document includes trapping support."""

        if self._pdf_trapped_raw == "True":
            return TrappedState.Yes
        elif self._pdf_trapped_raw == "False":
            return TrappedState.No

        return TrappedState.Unknown

    @pdf_trapped.setter
    def pdf_trapped(self, value: TrappedState) -> None:
        if value == TrappedState.Yes:
            self._pdf_trapped_raw = "True"
        elif value == TrappedState.No:
            self._pdf_trapped_raw = "False"
        elif value == TrappedState.Unknown:
            try:
                del self._pdf_trapped_raw
            except ValueError:
                pass

    @pdf_trapped.deleter
    def pdf_trapped(self) -> None:
        del self._pdf_trapped_raw

    # * XMP namespace properties
    # * https://developer.adobe.com/xmp/docs/XMPNamespaces/xmp/

    xmp_creator_tool = XMPTextProperty(namespaces["xmp"], "CreatorTool")
    """The name of the first known tool that created this resource."""

    xmp_create_date = XMPDateProperty(namespaces["xmp"], "CreateDate")
    """The datetime this resource was created. This need not match the file system
    creation date."""

    xmp_metadata_date = XMPDateProperty(namespaces["xmp"], "MetadataDate")
    """The datetime this metadata was last modified. It should be the same or more
    recent than :attr:`.modify_date`."""

    xmp_modify_date = XMPDateProperty(namespaces["xmp"], "ModifyDate")
    """The datetime this resource was last modified."""

    # * Dublin Core (DC) namespace properties
    # * https://developer.adobe.com/xmp/docs/XMPNamespaces/dc/

    dc_title = XMPLangAltProperty(namespaces["dc"], "title")
    """The titles or names given to this resource as a mapping of language names to titles."""

    dc_creator = XMPListProperty(namespaces["dc"], "creator", kind="Seq")
    """The entities primarily responsible for creating this resource."""

    dc_subject = XMPListProperty(namespaces["dc"], "subject", kind="Bag")
    """The topics or descriptions specifying the content of this resource."""

    dc_description = XMPLangAltProperty(namespaces["dc"], "description")
    """Textual descriptions of this resource as a mapping of language names to items."""

    dc_rights = XMPLangAltProperty(namespaces["dc"], "rights")
    """Rights statements pertaining to this resource."""

    dc_format = XMPTextProperty(namespaces["dc"], "format")
    """The MIME type of this resource."""



from __future__ import annotations

import datetime
import enum

from typing_extensions import Self

from ..common.fields import DateField, EnumField, PdfDictionary, TextStringField


class TrappedState(enum.Enum):
    """The document trapping support state."""

    No = 0
    """Document has not been trapped."""
    Yes = 1
    """Document has been trapped."""
    Unknown = 2
    """Unknown whether document is trapped partly, fully, or at all."""


class Info(PdfDictionary):
    """Document-level metadata representing the structure described in § 14.3.3,
    "Document information dictionary".

    Since PDF 2.0, most of its keys have been deprecated in favor of their equivalents
    in the document-level Metadata stream. The only keys not deprecated are the
    CreationDate and ModDate keys.
    """

    title = TextStringField("Title")
    """The document's title."""

    author = TextStringField("Author")
    """The name of the person who created the document."""

    subject = TextStringField("Subject")
    """The subject or topic of the document."""

    keywords = TextStringField("Keywords")
    """Keywords associated with the document."""

    creator = TextStringField("Creator")
    """If the document was converted to PDF from another format (ex. DOCX), the name of 
    the PDF processor that created the original document from which it was converted 
    (ex. Microsoft Word)."""

    producer = TextStringField("Producer")
    """If the document was converted to PDF from another format (ex. PostScript), the name of 
    the PDF processor that converted it to PDF (ex. Adobe Distiller)."""

    creation_date_raw = TextStringField("CreationDate")
    """The date and time the document was created, as a text string."""

    modify_date_raw = TextStringField("ModDate")
    """The date and time the document was most recently modified, as a text string."""

    creation_date = DateField("CreationDate")
    """The date and time the document was created, in human-readable form."""

    modify_date = DateField("ModDate")
    """The date and time the document was most recently modified, in human-readable form."""

    trapped = EnumField(
        "Trapped",
        {"True": TrappedState.Yes, "False": TrappedState.No, "Unknown": TrappedState.Unknown},
        TrappedState.Unknown,
    )
    """A value indicating whether the document has been modified to include trapping 
    information (see § 14.11.6, "Trapping support")."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls()
        dictionary.update()
        dictionary.data = mapping.data

        return dictionary

    def __init__(
        self,
        title: str | None = None,
        author: str | None = None,
        subject: str | None = None,
        keywords: str | None = None,
        creator: str | None = None,
        producer: str | None = None,
        creation_date: datetime.datetime | None = None,
        modify_date: datetime.datetime | None = None,
        trapped: TrappedState | None = None,
    ) -> None:
        super().__init__()

        # TODO: I'll rework this to be dataclassy some day.
        self._attrs = {
            "title": title,
            "author": author,
            "subject": subject,
            "keywords": keywords,
            "creator": creator,
            "producer": producer,
            "creation_date": creation_date,
            "modify_date": modify_date,
            "trapped": trapped,
        }

        for name, value in self._attrs.items():
            if value is not None:
                setattr(self, name, value)

    def __repr__(self) -> str:
        attributes = ", ".join(
            f"{attr}={value!r}"
            for attr in self._attrs
            if (value := getattr(self, attr, None)) is not None
        )

        return f"{self.__class__.__name__}({attributes})"



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from .containers import PdfDictionary


@dataclass
class PdfXRefSection:
    """A cross-reference section in a XRef table representing an incremental update.

    Each section is comprised of one or multiple subsections containing XRef entries.
    """

    subsections: list[PdfXRefSubsection]
    """The subsections conforming this XRef section."""

    trailer: PdfDictionary
    """The trailer dictionary specified within this XRef section."""


@dataclass
class PdfXRefSubsection:
    """A cross-reference subsection in an XRef section."""

    first_obj_number: int
    """The object number of the first entry in this section. Each entry's object number 
    starts here and is incremented by one."""

    count: int
    """The number of entries in this subsection."""

    entries: list[PdfXRefEntry]
    """The entries contained in this subsection."""


@dataclass
class FreeXRefEntry:
    """A Type 0 (``f``) or free entry. Free entries are entries not currently in use and are
    members of the linked list of free objects."""

    next_free_object: int
    """The object number of the next free object in the linked list."""

    gen_if_used_again: int
    """The generation to apply to an object if this entry is used again."""


@dataclass
class InUseXRefEntry:
    """A Type 1 (``n``) or in-use entry. In-use entries refer to the objects stored in a
    document."""

    offset: int
    """The byte offset of the object in the file (starting after the %PDF marker)."""

    generation: int
    """The generation of the object."""


@dataclass
class CompressedXRefEntry:
    """A Type 2 or compressed entry. Compressed entries refer to objects stored within
    an object stream."""

    objstm_number: int
    """The object number of the object stream containing this object."""

    index_within: int
    """The index of the object within the object stream."""


PdfXRefEntry = Union[FreeXRefEntry, InUseXRefEntry, CompressedXRefEntry]



from __future__ import annotations

from io import BytesIO

from pdfnaut import PdfDocument
from pdfnaut.cos.objects import PdfArray
from pdfnaut.objects import Page


def test_get_object() -> None:
    # Document with traditional xref table
    pdf = PdfDocument.from_filename(r"tests\docs\pdf2-incremental.pdf")

    assert pdf.objects[1] is pdf.catalog
    assert pdf.get_object((1, 0), cache=False) is not pdf.objects[1]

    # Document with compressed xref table
    pdf = PdfDocument.from_filename(r"tests\docs\compressed-xref.pdf")

    assert pdf.objects[1] is pdf.page_tree
    assert pdf.get_object((1, 0), cache=False) is not pdf.objects[1]


def test_insert_pages_to_new_doc() -> None:
    pdf = PdfDocument.new()

    page1 = Page(size=(595, 842))
    page1["Ident"] = b"It is me!"

    page2 = Page(size=(595, 842))
    page2["Ident"] = b"I'm another page!"

    pdf.pages.insert(0, page1)
    pdf.pages.insert(0, page2)

    assert pdf.pages[0]["Ident"] == page2["Ident"]
    assert pdf.pages[1]["Ident"] == page1["Ident"]


def test_add_pages_to_doc_with_flat_tree() -> None:
    origin_pdf = PdfDocument.from_filename(r"tests\docs\pdf2-incremental.pdf")

    origin_pdf.pages.append(Page(size=(500, 500)))
    origin_pdf.pages.insert(0, Page(size=(300, 300)))

    origin_pdf.save(docdata := BytesIO())

    new_pdf = PdfDocument(docdata.getvalue())

    assert len(new_pdf.pages) == 3
    assert new_pdf.pages[0].mediabox == PdfArray([0, 0, 300, 300])
    assert new_pdf.pages[-1].mediabox == PdfArray([0, 0, 500, 500])


def test_add_pages_to_doc_with_nested_tree() -> None:
    orig_pdf = PdfDocument.from_filename(r"tests\docs\pdf-with-page-tree.pdf")

    orig_pdf.pages.append(p1 := Page(size=(500, 500)))
    orig_pdf.pages.insert(0, p2 := Page(size=(300, 300)))

    assert p1.indirect_ref is not None and p2.indirect_ref is not None
    orig_pdf.save(docdata := BytesIO())

    # saved changes
    saved_pdf = PdfDocument(docdata.getvalue())
    assert len(saved_pdf.pages) == 6
    assert saved_pdf.pages[0].mediabox == PdfArray([0, 0, 300, 300])
    assert saved_pdf.pages[-1].mediabox == PdfArray([0, 0, 500, 500])


def test_remove_pages_from_doc() -> None:
    # flat tree
    pdf = PdfDocument.from_filename(r"tests\docs\pdf2-incremental.pdf")

    last_page = pdf.pages[-1]
    assert pdf.pages.pop() is last_page
    assert last_page.indirect_ref is None
    assert len(pdf.pages) == 0

    # nested tree
    pdf = PdfDocument.from_filename(r"tests\docs\pdf-with-page-tree.pdf")

    second_page = pdf.pages[1]
    assert pdf.pages.pop(1) is second_page
    assert second_page.indirect_ref is None
    assert len(pdf.pages) == 3

    # nested tree via delitem
    pdf = PdfDocument.from_filename(r"tests\docs\pdf-with-page-tree.pdf")

    second_page = pdf.pages[1]
    del pdf.pages[1]

    assert second_page.indirect_ref is None
    assert len(pdf.pages) == 3


def test_replace_page() -> None:
    pdf = PdfDocument.from_filename("tests/docs/pdf-with-page-tree.pdf")

    prev_page = pdf.pages[0]
    new_page = Page(size=(612.4, 791))
    new_page["Ident"] = b"It is me!"

    pdf.pages[0] = new_page

    # check if the replacement invalidated the previous page's reference
    assert prev_page.indirect_ref is None
    assert new_page.indirect_ref is not None

    # check our work
    assert pdf.pages[0] is new_page


def test_replace_page_from_doc_to_doc() -> None:
    origin_pdf = PdfDocument.from_filename(r"tests\docs\pdf-with-page-tree.pdf")
    replacing_pdf = PdfDocument.from_filename(r"tests\docs\pdf2-incremental.pdf")

    origin_pdf.pages[0] = replacing_pdf.pages[0]

    replaced_page = origin_pdf.pages[0]
    source_page = replacing_pdf.pages[0]

    assert replaced_page.indirect_ref != source_page.indirect_ref
